from collections import defaultdict
from textwrap import dedent

from django.db import transaction
from drf_spectacular.utils import extend_schema_serializer
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.documents.models import Corpus, Entity, EntityType, TranscriptionEntity
from arkindex.documents.serializers.light import CorpusLightSerializer, EntityTypeLightSerializer
from arkindex.documents.serializers.ml import WorkerRunSummarySerializer
from arkindex.project.serializer_fields import ForbiddenField, WorkerRunIDField
from arkindex.project.validators import ForbiddenValidator


class EntityTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = EntityType
        fields = (
            "id",
            "name",
            "color",
            "corpus"
        )
        read_only_fields = (
            "id",
            "corpus"
        )

    def to_internal_value(self, data):
        # Get an error if a request tries to change an entity type's corpus
        corpus = data.get("corpus")
        if self.instance and corpus:
            raise ValidationError({"corpus": ["It is not possible to update an Entity Type's corpus."]})
        data = super().to_internal_value(data)
        return data

    def validate(self, data):
        name = data.get("name") if "name" in data else self.instance.name
        corpus = data.get("corpus") if "corpus" in data else self.instance.corpus
        existing_type = corpus.entity_types.filter(name=name).using("default").first()
        if existing_type is not None and (not self.instance or existing_type.id != self.instance.id):
            raise ValidationError({"name": [f"Entity Type with name {name} already exists in corpus {corpus}."]})
        return data


class EntityTypeCreateSerializer(EntityTypeSerializer):
    class Meta:
        model = EntityType
        fields = EntityTypeSerializer.Meta.fields
        validators = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get("request"):
            # Keep original fields so OpenAPI can build its schemas
            return
        self.fields["corpus"].queryset = Corpus.objects.admin(self.context["request"].user)


@extend_schema_serializer(deprecate_fields=("worker_version_id", ))
class BaseEntitySerializer(serializers.ModelSerializer):
    """
    Serializes an entity
    """
    type = EntityTypeLightSerializer()
    metas = serializers.HStoreField(child=serializers.CharField(), required=False, allow_null=True)
    worker_version_id = serializers.PrimaryKeyRelatedField(read_only=True)
    worker_run = WorkerRunSummarySerializer(read_only=True, allow_null=True)

    class Meta:
        model = Entity
        fields = (
            "id",
            "name",
            "type",
            "metas",
            "validated",
            "worker_version_id",
            "worker_run",
        )
        read_only_fields = (
            "id",
            "worker_version_id",
            "worker_run",
        )


@extend_schema_serializer(
    deprecate_fields=("worker_version_id")
)
class EntitySerializer(BaseEntitySerializer):
    """
    Serialize an entity with its metadata
    """
    corpus = CorpusLightSerializer(read_only=True)
    # When updating an entity, the type can be set either by using its EntityType UUID, or its name
    # (in which case the serializer checks that an EntityType with this name exists in the corpus)
    type_id = serializers.PrimaryKeyRelatedField(
        queryset=EntityType.objects.all(),
        style={"base_template": "input.html"},
        required=True,
        write_only=True
    )
    type = EntityTypeLightSerializer(read_only=True)

    class Meta:
        model = Entity
        fields = BaseEntitySerializer.Meta.fields + (
            "corpus",
            "type_id"
        )
        read_only_fields = BaseEntitySerializer.Meta.read_only_fields = (
            "corpus",
        )

    def validate(self, data):
        errors = defaultdict(list)

        if data.get("validated"):
            data["moderator"] = self.context["request"].user
        corpus = self.instance.corpus
        type = data.get("type_id") or self.instance.type
        if type.corpus_id != corpus.id:
            errors["type_id"].append(f"EntityType {type.id} does not exist in corpus {corpus.name}.")
        data["type"] = type

        if errors:
            raise ValidationError(errors)
        return data


@extend_schema_serializer(
    deprecate_fields=("worker_version", "worker_version_id")
)
class EntityCreateSerializer(BaseEntitySerializer):
    """
    Serialize an entity with possible parents and children
    """
    corpus = serializers.PrimaryKeyRelatedField(
        queryset=Corpus.objects.none(),
        style={"base_template": "input.html"},
    )
    metas = serializers.HStoreField(child=serializers.CharField(), required=False)
    worker_version = serializers.UUIDField(
        allow_null=True,
        required=False,
        source="worker_version_id",
        validators=[
            ForbiddenValidator(),
        ],
        help_text=dedent("""
            ID of a WorkerVersion that created this entity.

            Creating new entities with a WorkerVersion is forbidden. Use `worker_run_id` instead.
        """),
    )
    worker_run_id = WorkerRunIDField(
        default=None,
        help_text=dedent("""
            A WorkerRun ID that the new entity will refer to.

            Regular users may only use the WorkerRuns of their own `Local` process.

            Tasks authenticated via the Ponos task authentication may only use the WorkerRuns of their process.
        """).strip(),

    )
    type_id = serializers.PrimaryKeyRelatedField(
        queryset=EntityType.objects.all(),
        style={"base_template": "input.html"},
        required=True,
    )

    class Meta:
        model = Entity
        fields = (
            "id",
            "name",
            "type_id",
            "metas",
            "validated",
            "corpus",
            "worker_version",
            "worker_run_id"
        )
        read_only_fields = (
            "id",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get("request"):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context["request"].user)
        self.fields["corpus"].queryset = corpora

    def validate(self, data):
        errors = defaultdict(list)

        worker_run = data.get("worker_run")
        if worker_run:
            data["worker_version_id"] = worker_run.version_id

        corpus = data.get("corpus")
        type = data.get("type_id")
        if type.corpus_id != corpus.id:
            errors["type_id"].append(f"EntityType {type.id} does not exist in corpus {corpus.name}.")
        data["type"] = type

        if errors:
            raise ValidationError(errors)

        return data


class TranscriptionEntityCreateSerializer(serializers.ModelSerializer):
    """
    Serialise the link between an entity and a transcription
    Transcription element must be passed in serializer context
    """
    offset = serializers.IntegerField(min_value=0)
    length = serializers.IntegerField(min_value=1)
    worker_version_id = ForbiddenField()
    worker_run_id = WorkerRunIDField(
        required=False,
        write_only=True,
        allow_null=True,
        help_text=dedent("""
            A WorkerRun ID that the new transcription and entity will refer to.

            Regular users may only use the WorkerRuns of their own `Local` process.

            Tasks authenticated via the Ponos task authentication may only use the WorkerRuns of their process.
        """).strip(),

    )
    confidence = serializers.FloatField(min_value=0, max_value=1, default=None)

    worker_run = WorkerRunSummarySerializer(read_only=True, allow_null=True)

    class Meta:
        model = TranscriptionEntity
        fields = (
            "entity",
            "offset",
            "length",
            "worker_version_id",
            "worker_run_id",
            "worker_run",
            "confidence",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get("request"):
            # Do not raise Error in order to create OpenAPI schema
            return

        # We need to use the default database to avoid stale read
        # when assigning a newly created entity on on a newly created transcription
        self.fields["entity"].queryset = Entity.objects.using("default").filter(
            corpus__in=Corpus.objects.writable(self.context["request"].user)
        )

    def validate(self, data):
        data = super().validate(data)
        errors = {}

        data["transcription"] = self.context["transcription"]
        if not data["transcription"]:
            errors["transcription"] = ["invalid UUID or Corpus write-access is forbidden"]

        worker_run = data.get("worker_run")
        if worker_run is not None:
            data["worker_version_id"] = worker_run.version_id

        existing_transcription_entities = TranscriptionEntity.objects.filter(
            transcription=data["transcription"],
            entity=data["entity"],
            offset=data["offset"],
            length=data["length"],
            worker_run=worker_run,
        )
        if existing_transcription_entities.exists():
            if worker_run:
                errors["__all__"] = ["This entity is already linked to this transcription by this worker run at this position."]
            else:
                errors["__all__"] = ["This entity is already linked to this transcription at this position."]

        if errors:
            raise serializers.ValidationError(errors)

        return data


@extend_schema_serializer(deprecate_fields=("worker_version_id", ))
class TranscriptionEntitySerializer(TranscriptionEntityCreateSerializer):
    """
    Serialize the link between a transcription and a detailed entity
    """
    worker_version_id = serializers.UUIDField(read_only=True)
    entity = BaseEntitySerializer()


class TranscriptionEntityBulkItemSerializer(serializers.ModelSerializer):
    name = serializers.CharField(write_only=True)
    type_id = serializers.UUIDField(
        help_text="UUID of the EntityType to use for this entity.",
        write_only=True,
    )
    transcription_entity_id = serializers.IntegerField(
        help_text="ID of the newly created TranscriptionEntity.",
        read_only=True,
    )

    class Meta:
        model = TranscriptionEntity
        fields = (
            "name",
            "type_id",
            "offset",
            "length",
            "confidence",
            "transcription_entity_id",
            "entity_id",
        )
        read_only_fields = (
            "transcription_entity_id",
            "entity_id",
        )
        extra_kwargs = {
            "offset": {"write_only": True},
            "length": {"write_only": True},
            "confidence": {"write_only": True},
            "entity_id": {"help_text": "UUID of the newly created Entity."},
        }

    def validate(self, data):
        # Ensure no TranscriptionEntity overflows transcription's length
        if data["offset"] + data["length"] > len(self.context["transcription"].text):
            raise ValidationError({"__all__": ["Entity position overflows transcription text size"]})
        return data


class TranscriptionEntitiesBulkSerializer(serializers.Serializer):
    worker_run_id = WorkerRunIDField(
        write_only=True,
        help_text=dedent("""
            A WorkerRun ID that the new entities and links to the transcription will refer to.

            Regular users may only use the WorkerRuns of their own `Local` process.

            Tasks authenticated via the Ponos task authentication may only use the WorkerRuns of their process.
        """).strip(),

    )
    entities = TranscriptionEntityBulkItemSerializer(
        many=True,
        allow_empty=False,
        help_text="Attributes for the Entities and TranscriptionEntities to create. Each item must be unique.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transcription = self.context.get("transcription")
        # Pass the transcription through to the item serializer for validation
        self.fields["entities"].context["transcription"] = self.transcription

    def validate(self, data):
        errors = {}

        if self.transcription.transcription_entities.filter(worker_run=data["worker_run"]).exists():
            errors["worker_run_id"] = ["Some TranscriptionEntities were already created for this worker run on this transcription."]

        # Check that all the EntityType IDs exist in one query
        found_types = set(self.transcription.element.corpus.entity_types.filter(
            id__in={item["type_id"] for item in data["entities"]}
        ).values_list("id", flat=True))

        entity_errors = {}
        for i, item in enumerate(data["entities"]):
            if item["type_id"] not in found_types:
                entity_errors[i] = {"type_id": ["An EntityType with this ID does not exist in the transcription's corpus."]}

        # Verify all TranscriptionEntities will be unique: the DB unique constraint that applies here is
        # (transcription, entity, offset, length, worker_run), so here we need to check on (entity, offset, length).
        # Since we are creating entities based on the (name, type_id) combo, we use (name, type_id, offset, length).
        unique_count = len(set(
            (item["offset"], item["length"], item["name"], item["type_id"])
            for item in data["entities"]
        ))
        # If we have a different entity count with the set than in the original list, there are duplicates.
        if len(data["entities"]) != unique_count:
            entity_errors["__all__"] = ["Some TranscriptionEntities have the same name, type, offset and length."]

        if entity_errors:
            errors["entities"] = entity_errors

        if errors:
            raise ValidationError(errors)

        return data

    @transaction.atomic
    def save(self):
        entities = Entity.objects.bulk_create([
            Entity(
                corpus=self.transcription.element.corpus,
                name=item["name"],
                type_id=item["type_id"],
                validated=True,
                worker_run=self.validated_data["worker_run"],
                worker_version_id=self.validated_data["worker_run"].version_id,
            )
            for item in self.validated_data["entities"]
        ])

        transcription_entities = TranscriptionEntity.objects.bulk_create([
            TranscriptionEntity(
                transcription=self.transcription,
                entity=entity,
                offset=item["offset"],
                length=item["length"],
                confidence=item.get("confidence"),
                worker_run=self.validated_data["worker_run"],
                worker_version_id=self.validated_data["worker_run"].version_id,
            )
            for entity, item in zip(entities, self.validated_data["entities"])
        ])

        # Update self.validated_data so that DRF can output a response with the created UUIDs
        self.validated_data["entities"] = [
            {
                "entity_id": transcription_entity.entity.id,
                "transcription_entity_id": transcription_entity.id,
            }
            for transcription_entity in transcription_entities
        ]
