from django.db.models.signals import pre_delete

from arkindex.documents.models import Corpus, Element, EntityType, MetaType, Transcription
from arkindex.documents.tasks import corpus_delete
from arkindex.ponos.models import Farm, State, Task
from arkindex.process.models import (
    CorpusWorkerVersion,
    Process,
    ProcessDatasetSet,
    ProcessMode,
    Repository,
    WorkerVersion,
)
from arkindex.project.tests import FixtureTestCase, force_constraints_immediate
from arkindex.training.models import Dataset, DatasetSet


class TestDeleteCorpus(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        # Add some data to the test corpus to test for deletion
        element = cls.corpus.elements.get(name="Volume 1")
        cls.user.selected_elements.add(element)

        file_import = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Files,
            folder_type=cls.corpus.types.get(slug="volume"),
            farm=Farm.objects.first(),
        )
        file_import.files.create(
            corpus=cls.corpus,
            name="arcanine.jpg",
            size=1337,
            content_type="image/jpeg",
        )
        file_import_version = WorkerVersion.objects.get(worker__slug="file_import")
        with cls.settings(cls, IMPORTS_WORKER_VERSION=str(file_import_version.id)):
            file_import.run()

        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")

        element_process = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            ml_class=cls.corpus.ml_classes.create(name="a class"),
        )
        element_process.elements.add(element)
        worker_run = element_process.worker_runs.create(version=cls.worker_version, parents=[])
        task_1, task_2, task_3 = Task.objects.bulk_create(
            [
                Task(
                    run=0,
                    depth=0,
                    process=element_process,
                    worker_run=worker_run,
                    slug=f"unscheduled task {i}",
                    state=State.Unscheduled,
                ) for i in range(1, 4)
            ]
        )
        task_1.parents.set([task_2])
        task_3.parents.set([task_1, task_2])
        element.worker_run = worker_run
        element.worker_version = cls.worker_version
        element.save()

        element.classifications.create(
            ml_class=cls.corpus.ml_classes.create(name="something"),
            worker_run=worker_run,
            worker_version=cls.worker_version,
        )
        element.metadatas.create(
            type=MetaType.Text,
            name="meta",
            value="data",
            worker_run=worker_run,
            worker_version=cls.worker_version,
        )
        element.transcriptions.create(
            text="blah",
            confidence=0.99,
            worker_run=worker_run,
            worker_version=cls.worker_version,
        )

        person_type = EntityType.objects.get(name="person", corpus=cls.corpus)

        entity1 = cls.corpus.entities.create(name="Magnemite", type=person_type)

        Transcription.objects.filter(element__corpus=cls.corpus).first().transcription_entities.create(
            entity=entity1,
            offset=1,
            length=1,
            worker_version=cls.worker_version,
            worker_run=worker_run,
        )

        # Create a separate corpus that should not get anything deleted
        cls.repo = Repository.objects.create(
            url="http://lol.git",
        )
        cls.corpus2 = Corpus.objects.create(name="Other corpus")

        dataset1 = Dataset.objects.get(name="First Dataset")
        test_set_1 = dataset1.sets.get(name="test")
        test_set_1.set_elements.create(element=element)
        cls.dataset2 = Dataset.objects.create(name="Dead Sea Scrolls", description="How to trigger a Third Impact", creator=cls.user, corpus=cls.corpus2)
        DatasetSet.objects.bulk_create(
            DatasetSet(
                dataset=cls.dataset2,
                name=set_name
            ) for set_name in ["test", "training", "validation"]
        )
        # Process on cls.corpus and with a set from cls.corpus
        dataset_process1 = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Dataset)
        ProcessDatasetSet.objects.create(process=dataset_process1, set=test_set_1)
        # Process on cls.corpus with a set from another corpus
        dataset_process2 = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Dataset)
        ProcessDatasetSet.objects.create(process=dataset_process2, set=test_set_1)
        ProcessDatasetSet.objects.create(process=dataset_process2, set=cls.dataset2.sets.get(name="training"))
        # Process on another corpus with a set from another corpus and none from cls.corpus
        cls.dataset_process3 = cls.corpus2.processes.create(creator=cls.user, mode=ProcessMode.Dataset)
        ProcessDatasetSet.objects.create(process=cls.dataset_process3, set=cls.dataset2.sets.get(name="validation"))

        cls.rev = cls.repo.revisions.create(
            hash="42",
            message="oh",
            author="me",
        )
        cls.process = Process.objects.create(
            creator=cls.user,
            corpus=cls.corpus2,
            mode=ProcessMode.Files,
        )
        cls.df = cls.process.files.create(
            name="a.txt",
            size=1234,
            content_type="text/plain",
            corpus=cls.corpus2,
        )
        cls.vol = cls.corpus2.elements.create(
            type=cls.corpus2.types.create(slug="volume", folder=True),
            name="A volume",
        )
        cls.vol.metadatas.create(
            type=MetaType.Text,
            name="meta",
            value="data",
        )
        cls.page = cls.corpus2.elements.create(
            type=cls.corpus2.types.create(slug="page"),
            name="A page",
        )
        cls.page.classifications.create(
            worker_version=cls.worker_version,
            ml_class=cls.corpus2.ml_classes.create(
                name="klass",
            ),
            confidence=0.5,
        )
        cls.page.transcriptions.create(
            worker_version=cls.worker_version,
            text="hi",
            confidence=0.75,
        )
        cls.page.add_parent(cls.vol)

        cls.page.activities.create(
            worker_version=cls.worker_version,
            process=element_process,
        )

        # Generate some CorpusWorkerVersions automatically
        CorpusWorkerVersion.objects.rebuild()

    def test_run(self):
        receivers = pre_delete.receivers

        with force_constraints_immediate(), self.assertExactQueries("corpus_delete.sql", params={"corpus_id": self.corpus.id}):
            corpus_delete(self.corpus.id)

        # Ensure the task restores the signal receivers
        self.assertEqual(pre_delete.receivers, receivers)

        self.assertEqual(Corpus.objects.count(), 1)
        with self.assertRaises(Corpus.DoesNotExist):
            self.corpus.refresh_from_db()
        self.corpus2.refresh_from_db()
        self.assertEqual(self.corpus2.types.count(), 2)
        self.assertEqual(self.corpus2.elements.count(), 2)
        self.assertEqual(Element.objects.count(), 2)

        self.repo.refresh_from_db()
        self.rev.refresh_from_db()
        self.process.refresh_from_db()
        self.df.refresh_from_db()
        self.vol.refresh_from_db()
        self.page.refresh_from_db()
        self.dataset_process3.refresh_from_db()

        self.assertTrue(self.repo.revisions.filter(id=self.rev.id).exists())
        self.assertEqual(self.process.files.get(), self.df)
        self.assertTrue(Element.objects.get_descending(self.vol.id).filter(id=self.page.id).exists())
        self.assertTrue(self.corpus2.datasets.filter(id=self.dataset2.id).exists())
        self.assertTrue(self.corpus2.processes.filter(id=self.dataset_process3.id).exists())

        md = self.vol.metadatas.get()
        self.assertEqual(md.name, "meta")
        self.assertEqual(md.value, "data")

        cl = self.page.classifications.get()
        self.assertEqual(cl.worker_version, self.worker_version)
        self.assertEqual(cl.ml_class.name, "klass")
        self.assertEqual(cl.confidence, 0.5)

        ts = self.page.transcriptions.get()
        self.assertEqual(ts.worker_version, self.worker_version)
        self.assertEqual(ts.text, "hi")
        self.assertEqual(ts.confidence, 0.75)

        self.assertEqual(self.corpus2.worker_version_cache.count(), 1)

    def test_run_top_level_type(self):
        receivers = pre_delete.receivers

        self.corpus.top_level_type = self.corpus.types.first()
        self.corpus.save()

        with force_constraints_immediate(), self.assertExactQueries("corpus_delete_top_level_type.sql", params={"corpus_id": self.corpus.id}):
            corpus_delete(self.corpus.id)

        # Ensure the task restores the signal receivers
        self.assertEqual(pre_delete.receivers, receivers)
