import logging
from collections import defaultdict
from datetime import timedelta
from textwrap import dedent
from uuid import UUID

from django.conf import settings
from django.db import transaction
from django.db.models import (
    Avg,
    Count,
    DurationField,
    Exists,
    ExpressionWrapper,
    F,
    IntegerField,
    Max,
    Min,
    OuterRef,
    Q,
    Value,
)
from django.db.models.functions import Coalesce, Now
from django.db.models.query import Prefetch
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from drf_spectacular.utils import (
    OpenApiExample,
    OpenApiParameter,
    OpenApiResponse,
    PolymorphicProxySerializer,
    extend_schema,
    extend_schema_view,
    inline_serializer,
)
from rest_framework import permissions, serializers, status
from rest_framework.authentication import SessionAuthentication, TokenAuthentication
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    DestroyAPIView,
    GenericAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveDestroyAPIView,
    RetrieveUpdateAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.response import Response
from rest_framework.serializers import Serializer

from arkindex.documents.models import Corpus, Element, Selection
from arkindex.ponos.authentication import TaskAuthentication
from arkindex.ponos.models import FINAL_STATES, STATES_ORDERING, State, Task
from arkindex.ponos.permissions import IsTask, IsVerifiedOrAgent
from arkindex.process.models import (
    ActivityState,
    DataFile,
    GitRef,
    GitRefType,
    Process,
    ProcessDatasetSet,
    ProcessMode,
    Revision,
    Worker,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerRun,
    WorkerType,
    WorkerVersion,
)
from arkindex.process.serializers.files import DataFileCreateSerializer, DataFileSerializer
from arkindex.process.serializers.git import RevisionSerializer
from arkindex.process.serializers.imports import (
    ApplyProcessTemplateSerializer,
    CorpusProcessSerializer,
    CreateProcessTemplateSerializer,
    FilesProcessSerializer,
    ProcessDetailsSerializer,
    ProcessElementLightSerializer,
    ProcessElementSerializer,
    ProcessListSerializer,
    ProcessSerializer,
    StartProcessSerializer,
)
from arkindex.process.serializers.ingest import BucketSerializer, S3ImportSerializer
from arkindex.process.serializers.training import ProcessDatasetSetSerializer
from arkindex.process.serializers.worker_runs import (
    CorpusWorkerRunSerializer,
    UserWorkerRunSerializer,
    WorkerRunEditSerializer,
    WorkerRunSerializer,
)
from arkindex.process.serializers.workers import (
    CorpusWorkerVersionSerializer,
    DockerWorkerVersionSerializer,
    ExistingWorkerException,
    RepositorySerializer,
    WorkerActivitySerializer,
    WorkerConfigurationExistsErrorSerializer,
    WorkerConfigurationListSerializer,
    WorkerConfigurationSerializer,
    WorkerCreateSerializer,
    WorkerSerializer,
    WorkerStatisticsSerializer,
    WorkerTypeSerializer,
    WorkerVersionCreateSerializer,
    WorkerVersionSerializer,
)
from arkindex.process.utils import annotate_image_url
from arkindex.project.aws import get_ingest_resource
from arkindex.project.fields import ArrayRemove
from arkindex.project.mixins import (
    ConflictAPIException,
    CorpusACLMixin,
    ProcessACLMixin,
    RepositoryACLMixin,
    SelectionMixin,
    WorkerACLMixin,
)
from arkindex.project.openapi import UUID_OR_STR
from arkindex.project.pagination import CountCursorPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.project.tools import PercentileCont
from arkindex.project.triggers import process_delete
from arkindex.training.models import DatasetSet, Model
from arkindex.users.models import Role, Scope

logger = logging.getLogger(__name__)


@extend_schema_view(
    get=extend_schema(
        operation_id="ListProcesses",
        tags=["process"],
        parameters=[
            OpenApiParameter(
                "corpus",
                type=UUID,
                description="Filter processes by corpus ID",
                required=False,
            ),
            OpenApiParameter(
                "created",
                type=bool,
                description="Restrict to or exclude processes you created",
                required=False,
            ),
            OpenApiParameter(
                "mode",
                enum=[mode.value for mode in ProcessMode],
                description="Filter processes by mode. When this filter is unset, the `local` mode is excluded.",
                required=False,
            ),
            OpenApiParameter(
                "state",
                enum=[state.value for state in State],
                description="Filter processes by state",
                required=False,
            ),
            OpenApiParameter(
                "id",
                description="Filter processes by beginning of UUID",
                required=False,
            ),
            OpenApiParameter(
                "name",
                description="Filter processes whose name contains the given string (case insensitive)",
                required=False,
            ),
            OpenApiParameter(
                "started",
                type=bool,
                description="Restrict to or exclude processes that have already been started. "
                            "This has no effect when filtering by the `template` or `local` modes, as those cannot start.",
                required=False,
            )
        ]
    )
)
class ProcessList(ProcessACLMixin, ListAPIView):
    """
    List all visible processes.
    """
    permission_classes = (IsVerified, )
    serializer_class = ProcessListSerializer

    def get_queryset(self):
        filters = Q()

        if "corpus" in self.request.query_params:
            corpus_id = self.request.query_params["corpus"]
            try:
                corpus_id = UUID(corpus_id)
            except (AttributeError, ValueError):
                raise ValidationError({"corpus": ["'{}' is not a valid UUID".format(corpus_id)]})
            # No supplementary validation is required on the corpus ID filter
            filters &= Q(corpus=corpus_id)

        # Default process_mode variable for compatibility later
        process_mode = None
        if "mode" in self.request.query_params:
            try:
                process_mode = ProcessMode(self.request.query_params["mode"])
                filters &= Q(mode=process_mode)
            except ValueError:
                raise ValidationError({
                    "mode": ["Mode '{}' does not exist".format(self.request.query_params["mode"])]
                })
        else:
            # Exclude local processes by default
            filters &= ~Q(mode=ProcessMode.Local)

        # When listing template or local processes, the task filter makes no sense
        has_started = None
        if process_mode not in (ProcessMode.Template, ProcessMode.Local) and "started" in self.request.query_params:
            has_started = self.request.query_params["started"].lower() not in ("false", "0")
            filters &= Q(started__isnull=not has_started)

        if "created" in self.request.query_params:
            created = self.request.query_params["created"]
            creator_filter = Q(creator_id=self.request.user.id)
            if created.lower() in ("false", "0"):
                creator_filter = ~creator_filter
            filters &= creator_filter

        if "id" in self.request.query_params:
            filters &= Q(id__startswith=self.request.query_params["id"])

        if "name" in self.request.query_params:
            filters &= Q(name__icontains=self.request.query_params["name"])

        qs = (
            self.readable_processes
            .select_related("creator")
            .filter(filters)
            # Order processes by completion date when available, or start date, or last update
            .annotate(date_order=Coalesce("finished", "started", "updated"))
        )

        state_value = self.request.query_params.get("state")

        if has_started is False or process_mode in (ProcessMode.Local, ProcessMode.Template):
            # If we are filtering to only include processes that have not started, or process modes that never have tasks,
            # then it is pointless to retrieve the tasks, and we know there is no last_run.
            qs = qs.annotate(last_run=Value(None, output_field=IntegerField()))
        else:
            # Otherwise, fetch all tasks.
            qs = qs.prefetch_related("tasks")
            # Only define `last_run` if we filter by state, since we will need it to do the filtering.
            # `last_run` is otherwise unnecessary, since it can be computed from all the prefetched tasks.
            if state_value:
                qs = qs.annotate(last_run=Max("tasks__run"))

        if state_value:
            try:
                state = State(self.request.query_params["state"])
            except ValueError:
                raise ValidationError({"state": [f"State '{state_value}' does not exist"]})

            # Filter out processes which have a task with an incompatible state on their last run
            excluding_states = STATES_ORDERING[:STATES_ORDERING.index(state)]
            excluded_processes = qs.filter(
                Q(tasks__run=F("last_run")),
                Q(tasks__state__in=excluding_states)
            )
            # Keep non excluded processes matching the state on their last run tasks
            state_query = Q(tasks__run=F("last_run")) & Q(tasks__state=state)
            if state == State.Unscheduled:
                # Handle the absence of tasks as unscheduled
                state_query |= Q(tasks__isnull=True)
            qs = qs.filter(state_query).exclude(id__in=excluded_processes.values("id")).distinct()

        return qs.order_by("-date_order")


class ProcessQuerysetMixin(object):
    """
    Optimized queryset for Retrieve/Update/PartialUpdate/Destroy/RetryProcess
    """

    def get_queryset(self):
        queryset = (
            Process
            .objects
            # Needed for ACL checks
            .select_related("corpus")
            # Needed for Process.state
            .prefetch_related("tasks")
            .annotate(last_run=Max("tasks__run"))
        )

        # When not deleting a process, we will be returning the whole process again, so we need to fetch from more tables
        if self.request.method.lower() != "delete":
            queryset = (
                queryset
                .select_related(
                    "corpus",
                    "farm",
                    # When an element is set on the process, it is serialized with its image URL and type
                    "element__image__server",
                    "element__type",
                    # Element and folder types are serialized as their slugs
                    "element_type",
                    "folder_type",
                )
                # Files and tasks are also listed
                .prefetch_related("files", "tasks__parents")
            )

        return queryset


@extend_schema(tags=["process"])
@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveProcess",
        description=dedent("""
            Retrieve details of a process.

            Requires a **guest** access.
        """),
    ),
    patch=extend_schema(
        operation_id="PartialUpdateProcess",
        description=dedent("""
            Partially update a process.

            Requires an **admin** access.
        """),
    ),
    put=extend_schema(
        operation_id="UpdateProcess",
        description=dedent("""
            Update a process.

            Requires an **admin** access.
        """),
    ),
    delete=extend_schema(
        operation_id="DestroyProcess",
        description=dedent("""
            Delete a process.

            Cannot be used on currently running processes. Requires an **admin** access to the process.

            If the process has worker activities, an asynchronous deletion is triggered and **HTTP 202 Accepted** is returned.

            Otherwise, the process is deleted immediately and HTTP 204 is returned.
        """),
        responses={
            202: None,
            204: None,
        },
    ),
)
class ProcessDetails(ProcessACLMixin, ProcessQuerysetMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve a process
    """
    permission_classes = (IsVerified, )
    serializer_class = ProcessDetailsSerializer

    def get_object(self):
        if not hasattr(self, "_process"):
            self._process = super().get_object()
        return self._process

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and "pk" in self.kwargs:
            context["process"] = self.get_object()
        return context

    def check_object_permissions(self, request, process):
        super().check_object_permissions(request, process)

        access_level = self.process_access_level(process)
        required_access = Role.Admin.value
        if request.method in permissions.SAFE_METHODS:
            # A guest role allow to retrieve information on a process
            required_access = Role.Guest.value

        if not access_level:
            raise NotFound
        if access_level < required_access:
            raise PermissionDenied(detail="You do not have a sufficient access level to this process.")

        if request.method == "DELETE" and process.state == State.Running:
            raise ValidationError({"__all__": ["Cannot delete a process while it is running"]})

    def destroy(self, request, *args, **kwargs):
        # If the process has no activity, deletes it directly. Triggers an async deletion otherwise.
        instance = self.get_object()
        if instance.activity_state == ActivityState.Pending:
            raise ValidationError({"__all__": ["Please wait activities to be initialized before deleting this process"]})
        if instance.worker_runs.all().in_use():
            raise ValidationError({
                "__all__": ["Some worker runs on this process are linked to ML results and cannot be deleted."]
            })
        if instance.activities.exists():
            process_delete(instance, delete_activity=WorkerActivityState.Queued)
            return Response(status=status.HTTP_202_ACCEPTED)
        instance.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


class ProcessRetry(ProcessACLMixin, ProcessQuerysetMixin, GenericAPIView):
    """
    Retry a process. Can only be used on processes with Error, Failed, Stopped or Completed states.\n\n
    Requires an **admin** access to the process and **guest** access to the process' farm.
    """
    permission_classes = (IsVerified, )
    serializer_class = ProcessSerializer

    def check_object_permissions(self, request, process):
        super().check_object_permissions(request, process)

        access_level = self.process_access_level(process)
        if not access_level:
            raise NotFound
        if access_level < Role.Admin.value:
            raise PermissionDenied(detail="You do not have an admin access to this process.")

        # When process.farm is None, the ProcessBuilder will already return an error, so we can ignore that case
        if process.farm and not process.farm.is_available(self.request.user):
            raise ValidationError({"farm": ["You do not have access to this farm."]})

        # process.state can cause new SQL queries to be run, so we access it just once
        state = process.state
        # Allow 'retrying' a process that has no Ponos tasks (that has never been started)
        if len(process.tasks.all()) and state in (State.Unscheduled, State.Pending):
            raise ValidationError({"__all__": ["This process is already pending"]})
        elif state == State.Running:
            raise ValidationError({"__all__": ["This process is already running"]})
        elif state == State.Stopping:
            raise ValidationError({"__all__": ["This process is stopping"]})

    @extend_schema(
        operation_id="RetryProcess",
        tags=["process"],
        request=None,
    )
    def post(self, request, *args, **kwargs):
        process = self.get_object()
        process.run()
        serializer = self.get_serializer(process)
        return Response(serializer.data)


@extend_schema(
    responses={201: ProcessSerializer},
    tags=["files"],
)
class FilesProcess(CreateAPIView):
    """
    Start a process from one or more uploaded files.
    """

    permission_classes = (IsVerified, )
    serializer_class = FilesProcessSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(
            status=status.HTTP_201_CREATED,
            data=ProcessSerializer(serializer.instance, context={"request": self.request}).data,
            headers=headers,
        )


@extend_schema_view(
    post=extend_schema(
        operation_id="CreateProcess",
        tags=["process"],
        responses={201: ProcessSerializer}
    )
)
class CorpusProcess(SelectionMixin, CorpusACLMixin, CreateAPIView):
    """
    Create a distributed process from elements of an Arkindex corpus.\n\n
    Requires an **admin** access to the corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = CorpusProcessSerializer

    @transaction.atomic
    def create(self, request, pk=None, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        corpus, element, process_name, name_contains, element_type, ml_class, selection, load_children, use_cache, mode = map(
            lambda key: serializer.validated_data.pop(key, None),
            (
                "corpus",
                "element",
                "name",
                "element_name_contains",
                "element_type",
                "ml_class",
                "selection",
                "load_children",
                "use_cache",
                "mode"
            )
        )

        process = corpus.processes.create(
            creator=self.request.user,
            mode=ProcessMode(mode),
            name=process_name,
            name_contains=name_contains,
            element_type=element_type,
            element=element,
            ml_class=ml_class,
            load_children=load_children,
            use_cache=use_cache,
        )

        # Ensure process elements do exist
        if selection:
            elements = self.get_selection(corpus.id)
            if not elements.exists():
                raise ValidationError({
                    "__all__": ["No element match those filters."]
                })
            process.elements.set(elements)

        return Response(
            status=status.HTTP_201_CREATED,
            data=ProcessSerializer(
                process,
                context=self.get_serializer_context()
            ).data,
        )


@extend_schema_view(
    post=extend_schema(
        operation_id="StartProcess",
        tags=["process"],
        responses=ProcessSerializer,
    ),
)
class StartProcess(CorpusACLMixin, CreateAPIView):
    """
    Start a process, used to build a Workflow with Workers.\n\n
    Requires an **admin** access to the corpus of this process.
    """
    permission_classes = (IsVerified, )
    serializer_class = StartProcessSerializer
    queryset = (
        Process
        .objects
        .select_related("corpus")
        .filter(corpus_id__isnull=False)
        .prefetch_related(Prefetch("worker_runs", queryset=WorkerRun.objects.select_related(
            "version__worker__repository",
            "model_version__model",
            "configuration",
        )))
        .prefetch_related(Prefetch("sets", queryset=DatasetSet.objects.select_related("dataset")))
        # Uses Exists() for has_tasks and not a __isnull because we are not joining on tasks and do not need to fetch them
        .annotate(has_tasks=Exists(Task.objects.filter(process=OuterRef("pk"))))
    )

    def check_object_permissions(self, request, process):
        super().check_object_permissions(request, process)

        if not self.has_admin_access(process.corpus):
            raise PermissionDenied(detail="You do not have an admin access to the corpus of this process.")

        if process.mode not in (ProcessMode.Workers, ProcessMode.Dataset) or process.has_tasks:
            raise ValidationError(
                {"__all__": ["Only a Process with Workers or Dataset mode and not already launched can be started later on"]})

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data, instance=self.get_object())
        serializer.is_valid(raise_exception=True)

        self.perform_create(serializer)

        headers = self.get_success_headers(serializer.data)
        response_serializer = ProcessSerializer(
            serializer.instance,
            context={"request": request},
        )

        return Response(
            response_serializer.data,
            status=status.HTTP_201_CREATED,
            headers=headers,
        )


@extend_schema(tags=["files"])
@extend_schema_view(
    get=extend_schema(
        operation_id="ListDataFile",
        parameters=[
            OpenApiParameter(
                "id",
                location=OpenApiParameter.PATH,
                type=UUID,
                description="ID of the corpus to list files on.",
            )
        ],
    ),
)
class DataFileList(CorpusACLMixin, ListAPIView):
    """
    List uploaded files in a corpus.

    Requires **guest** access to the corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer
    # For OpenAPI type discovery
    queryset = DataFile.objects.none()

    def get_queryset(self):
        return DataFile.objects.filter(corpus=self.get_corpus(self.kwargs["pk"]), trashed=False)


@extend_schema(tags=["files"])
@extend_schema_view(
    get=extend_schema(description="Get an uploaded file's metadata"),
    patch=extend_schema(description="Update a datafile's status"),
    put=extend_schema(description="Update a datafile's status"),
    delete=extend_schema(description="Delete an uploaded file")
)
class DataFileRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        # We need to use the default database to avoid stale read
        # when a datafile is updated immediately after it's created on a small upload
        return DataFile.objects.using("default").filter(
            corpus__in=Corpus.objects.readable(self.request.user)
        )

    def perform_update(self, serializer):
        if not self.has_write_access(serializer.instance.corpus):
            raise PermissionDenied
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied

        # The file is simply marked as trashed here and will be deleted asynchronously through a cron job
        instance.trashed = True
        instance.save()


@extend_schema_view(post=extend_schema(operation_id="CreateDataFile", tags=["files"]))
class DataFileCreate(CreateAPIView):
    """
    Create a Datafile.
    In case of success, a signed URI is returned to upload file content directly to a remote server.
    """

    permission_classes = (IsVerified, )
    serializer_class = DataFileCreateSerializer


@extend_schema(tags=["process"])
@extend_schema_view(
    get=extend_schema(
        operation_id="ListProcessSets",
        description=dedent(
            """
            List all dataset sets on a process.

            Requires a **guest** access to the process.
            """
        ),
    ),
)
class ProcessDatasetSets(ProcessACLMixin, ListAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ProcessDatasetSetSerializer
    queryset = ProcessDatasetSet.objects.none()

    @cached_property
    def process(self):
        process = get_object_or_404(
            Process.objects.using("default").select_related("corpus"),
            Q(pk=self.kwargs["pk"])
        )
        if not self.process_access_level(process):
            raise PermissionDenied(detail="You do not have guest access to this process.")
        return process

    def get_queryset(self):
        return (
            ProcessDatasetSet.objects.filter(process_id=self.process.id)
            .select_related("process__creator", "set__dataset__creator")
            .prefetch_related(Prefetch("set__dataset__sets", queryset=DatasetSet.objects.order_by("name")))
            .order_by("set__dataset__name", "set__name")
        )

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Ignore this step when generating the schema with OpenAPI
        if not self.kwargs:
            return context
        context["process"] = self.process
        return context


@extend_schema(tags=["process"])
@extend_schema_view(
    post=extend_schema(
        operation_id="CreateProcessSet",
        description=dedent(
            """
            Add a dataset set to a process.

            Requires an **admin** access to the process and a **guest** access to the dataset's corpus.
            """
        ),
    ),
    delete=extend_schema(
        operation_id="DestroyProcessSet",
        description=dedent(
            """
            Remove a dataset set from a process.

            Requires an **admin** access to the process.
            """
        ),
    ),
)
class ProcessDatasetSetManage(CreateAPIView, DestroyAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ProcessDatasetSetSerializer

    def get_object(self):
        qs = (
            ProcessDatasetSet.objects
            .select_related("set__dataset__creator", "process__corpus")
            # Required to check for a process that have already started
            .annotate(process_has_tasks=Exists(Task.objects.filter(process_id=self.kwargs["process"])))
        )
        # Only prefetch the dataset sets when creating
        if self.request.method != "DELETE":
            qs.prefetch_related(Prefetch("set__dataset__sets", queryset=DatasetSet.objects.order_by("name")))
        process_set = get_object_or_404(
            qs,
            set_id=self.kwargs["set"], process_id=self.kwargs["process"]
        )
        # Copy the has_tasks annotation onto the process
        process_set.process.has_tasks = process_set.process_has_tasks
        return process_set

    def destroy(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        get_object_or_404(ProcessDatasetSet, **serializer.validated_data).delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema_view(
    get=extend_schema(
        tags=["repos"],
        operation_id="ListRepositories",
        description="List connected repositories",
    )
)
class RepositoryList(RepositoryACLMixin, ListAPIView):
    """
    List connected repositories
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return self.readable_repositories \
            .annotate(authorized_users=Count("memberships")) \
            .prefetch_related("workers__type") \
            .order_by("url")


@extend_schema(
    tags=["repos"],
    description="Retrieve a repository",
)
@extend_schema_view(
    delete=extend_schema(description="Delete a repository")
)
class RepositoryRetrieve(RepositoryACLMixin, RetrieveDestroyAPIView):
    """
    Retrieve a repository
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return self.readable_repositories.annotate(authorized_users=Count("memberships"))

    def check_object_permissions(self, request, repo):
        if self.request.method not in permissions.SAFE_METHODS and not self.has_admin_access(repo):
            raise PermissionDenied(detail="You do not have admin access to this repository.")


@extend_schema(
    tags=["repos"],
    description="Retrieve a revision",
)
class RevisionRetrieve(RepositoryACLMixin, RetrieveAPIView):
    """
    Retrieve a revision
    """
    permission_classes = (IsVerified, )
    serializer_class = RevisionSerializer

    def get_queryset(self):
        return Revision.objects.filter(repo__in=self.executable_repositories)


@extend_schema(tags=["repos"])
@extend_schema_view(
    get=extend_schema(
        description="List workers for which the user has an execution access.",
        parameters=[
            OpenApiParameter(
                "repository_id",
                type=UUID,
                description="Filter workers by repository ID.",
                required=False,
            ),
            OpenApiParameter(
                "name",
                type=str,
                description="Filter workers by name.",
                required=False,
            ),
            OpenApiParameter(
                "type",
                type=UUID_OR_STR,
                description="Filter workers by type, using either a type UUID or a type slug.",
                required=False,
            ),
            OpenApiParameter(
                "archived",
                type=bool,
                description="Filter by or exclude workers that have been marked as archived.",
                required=False,
            ),
            OpenApiParameter(
                "compatible_model",
                type=UUID,
                description="Filter workers compatible with a specific model.",
                required=False,
            ),
        ]
    ),
    post=extend_schema(
        responses={
            201: WorkerSerializer,
            200: WorkerSerializer,
        },
        description=dedent("""
            Create a worker.
            Create or reuse a worker type based on the `type` value.

            Authentication can be done:
            * Using a user authentication (via a cookie or token).
            The user will be granted an administrator access to the worker in case it is created.
            * Using a ponos task authentication, linking the worker to the repository of the current process.
            The process must be of type `Repository`, have a revision and its creator must have an **admin**
            access to its repository.

            If a worker already exists with the given slug and the user has an execution access to it, an
            HTTP 200 is returned serializing its content.
        """),
    ),
)
class WorkerList(ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerCreateSerializer
    queryset = Worker.objects.none()

    def filter_queryset(self, qs):
        filters = Q()
        errors = defaultdict(list)

        compatible_model_id = self.request.query_params.get("compatible_model")
        if compatible_model_id:
            try:
                compatible_model_id = UUID(compatible_model_id)
                compatible_model = Model.objects.readable(self.request.user).get(id=compatible_model_id)
            except (TypeError, ValueError):
                errors["compatible_model"].append("Invalid UUID")
            except Model.DoesNotExist:
                errors["compatible_model"].append("This model does not exist or you don't have access to it")
            else:
                filters &= Q(models=compatible_model)

        repo_id = self.request.query_params.get("repository_id")
        if repo_id:
            try:
                repo_id = UUID(repo_id)
            except (TypeError, ValueError):
                errors["repository_id"].append("Invalid UUID")
            else:
                filters &= Q(repository_id=repo_id)

        name_filter = self.request.query_params.get("name")
        if name_filter:
            filters &= Q(name__icontains=name_filter)

        worker_type = self.request.query_params.get("type")
        if worker_type:
            try:
                worker_type_id = UUID(worker_type)
                filters &= Q(type_id=worker_type_id)
            except ValueError:
                # If the given type is not a worker_type id, it might be a worker_type__slug
                try:
                    worker_type = WorkerType.objects.get(slug=worker_type)
                except WorkerType.DoesNotExist:
                    errors["type"].append("No registered worker type with that slug.")
                else:
                    filters &= Q(type_id=worker_type.id)

        if "archived" in self.request.query_params:
            filters &= Q(archived__isnull=self.request.query_params["archived"].lower().strip() in ("false", "0"))

        if errors:
            raise ValidationError(errors)

        return super().filter_queryset(qs.filter(filters))

    def get_queryset(self):
        """
        List executable workers (i.e. via a direct write right or a write right on its repository).
        """
        return Worker.objects.executable(self.request.user) \
            .order_by("name", "id") \
            .select_related("type")

    def create(self, *args, **kwargs):
        try:
            return super().create(*args, **kwargs)
        except ExistingWorkerException as e:
            return Response(WorkerSerializer(e.worker).data, status=status.HTTP_200_OK)


@extend_schema(
    tags=["repos"],
    description="List available worker types on an Arkindex instance.",
)
class WorkerTypesList(ListAPIView):
    """
    List available worker types on instance
    """
    serializer_class = WorkerTypeSerializer
    queryset = WorkerType.objects.all().order_by("display_name")


@extend_schema(tags=["repos"])
@extend_schema_view(
    get=extend_schema(
        description=(
            "List versions for a given worker ID with their revision and associated git references.\n\n"
            "Requires an **execution** access to the worker."
        ),
        parameters=[
            OpenApiParameter(
                "mode",
                type=str,
                default="complete",
                description=("If set to `simple`, exclude versions other than master/main or with no tag")
            )
        ],
    ),
    post=extend_schema(
        description=dedent("""
            Create a new version for a worker.

            The user must have an administrator access to the worker.
        """)
    )
)
class WorkerVersionList(WorkerACLMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerVersionCreateSerializer
    queryset = WorkerVersion.objects.none()

    @property
    def simple_mode(self):
        return self.request.query_params.get("mode", "complete").lower() == "simple"

    @cached_property
    def worker(self):
        worker = get_object_or_404(
            Worker.objects.select_related("repository"),
            pk=self.kwargs["pk"]
        )
        if self.request.method in permissions.SAFE_METHODS and not self.has_execution_access(worker):
            raise PermissionDenied(detail="You do not have an execution access to this worker.")
        if (
            self.request.method not in permissions.SAFE_METHODS
            # An admin access is required for creation
            and not self.has_admin_access(worker)
        ):
            raise PermissionDenied(detail="You do not have an admin access to this worker.")
        return worker

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["worker"] = self.worker
        return context

    def get_queryset(self):
        filters = Q()
        if self.simple_mode:
            # Limit output to versions with tags or master/main branches
            filters = (
                Q(revision__refs__type=GitRefType.Tag)
                | Q(revision__refs__type=GitRefType.Branch, revision__refs__name__in=["master", "main"])
            )
        return self.worker.versions \
            .filter(filters) \
            .prefetch_related("revision__repo", "revision__refs", "worker__type") \
            .distinct() \
            .order_by("-revision__created")


@extend_schema_view(
    get=extend_schema(
        operation_id="ListCorpusWorkerVersions",
        tags=["ml"],
    )
)
class CorpusWorkerVersionList(CorpusACLMixin, ListAPIView):
    """
    List worker versions used by any ML result in a given corpus.

    Guest access is required on private corpora. No specific rights are required on the workers.
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = CorpusWorkerVersionSerializer
    # For OpenAPI type discovery
    queryset = WorkerVersion.objects.none()

    @cached_property
    def corpus(self):
        return get_object_or_404(Corpus.objects.readable(self.request.user), pk=self.kwargs["pk"])

    def get_queryset(self):
        return (
            self.corpus.worker_version_cache
            .select_related(
                "worker_version__worker__type",
                "worker_version__worker__repository",
                "worker_version__revision__repo",
                "model_version__model",
                "worker_configuration",
            )
            .order_by(
                "worker_version__worker__name",
                "worker_version__revision__hash",
                "model_version__model__name",
                "worker_configuration__name",
                "id",
            )
            .prefetch_related(
                "worker_version__revision__refs",
            )
        )


@extend_schema_view(
    get=extend_schema(
        operation_id="ListCorpusWorkerRuns",
        tags=["ml"],
    )
)
class CorpusWorkerRunList(ListAPIView):
    """
    List worker runs used by any ML result in a given corpus.

    The results also include all of the current user's worker runs that have results on any corpora, and not
    just the request's corpus, as user worker runs are independent from corpora.
    The results do **not** include other users' worker runs, even if they have produced results on the
    request's corpus.

    Guest access is required on private corpora. No specific rights are required on the workers.
    """

    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = CorpusWorkerRunSerializer
    # For OpenAPI type discovery
    queryset = WorkerRun.objects.none()

    @cached_property
    def corpus(self):
        return get_object_or_404(Corpus.objects.readable(self.request.user), pk=self.kwargs["pk"])

    def get_queryset(self):
        return (
            WorkerRun.objects.select_related("process__corpus")
            .select_related(
                "version__worker__type",
                "version__worker__repository",
                "version__revision__repo",
                "configuration",
                "model_version__model",
            )
            .prefetch_related(
                Prefetch(
                    "version__revision__refs",
                    queryset=GitRef.objects.select_related("repository")
                ),
                "process__tasks",
            )
            .filter(Q(process__corpus_id=self.corpus.id) | Q(process__creator_id=self.request.user.id, process__mode=ProcessMode.Local), has_results=True)
            .annotate(process_element_count=Count("process__elements"))
            .order_by("summary")
        )


@extend_schema(tags=["repos"])
@extend_schema_view(
    get=extend_schema(
        description=dedent("""
            Retrieve a worker version.
        """),
    ),
    patch=extend_schema(
        description=dedent("""
            Partially update a worker version.

            Requires authentication as a Ponos task.
        """),
    ),
    put=extend_schema(
        description=dedent("""
            Update a worker version.

            Requires authentication as a Ponos task.
        """),
    ),
)
class WorkerVersionRetrieve(RetrieveUpdateAPIView):
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = WorkerVersionSerializer
    queryset = WorkerVersion.objects.select_related("worker")

    def check_object_permissions(self, request, instance):
        """
        Allow any user to retrieve a worker version without authentication
        This behavior allows to retrieve a version information on public resources

        Only Ponos tasks are allowed to update a worker version
        """
        if self.request.method in permissions.SAFE_METHODS:
            return
        super().check_object_permissions(request, instance)
        if not isinstance(self.request.auth, Task):
            raise PermissionDenied(detail="Only Ponos tasks are allowed to update a worker version.")


@extend_schema(tags=["repos"])
@extend_schema_view(
    get=extend_schema(
        description=dedent("""
            Retrieve a worker.

            Requires a contributor access to the worker or its repository, except for `public` workers.
        """),
    ),
    put=extend_schema(
        description=dedent("""
            Update a worker.

            Requires a contributor access to the worker or its repository.
        """),
    ),
    patch=extend_schema(
        description=dedent("""
            Partially update a worker.

            Requires a contributor access to the worker or its repository.
        """),
    ),
)
class WorkerRetrieve(RetrieveUpdateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerSerializer
    queryset = Worker.objects.none()

    def get_queryset(self):
        return Worker.objects.executable(self.request.user).select_related("repository", "type")

    def check_object_permissions(self, request, worker):
        if self.request.method not in permissions.SAFE_METHODS and not worker.is_editable(self.request.user):
            raise PermissionDenied(detail="You do not have a contributor access to this worker.")


@extend_schema(tags=["repos"])
@extend_schema_view(
    get=extend_schema(
        description=(
            "List configurations for a given worker ID.\n\n"
            "Requires a **read** access to the worker or its repository."
        ),
        parameters=[
            OpenApiParameter(
                "archived",
                type=bool,
                description="Filter configurations by their `archived` state. "
                            "By default, the list only displays enabled (non-archived) configurations.",
                required=False,
                default=False,
            )
        ]
    ),
    post=extend_schema(
        description=(
            "Create a configuration for a given worker ID, name and JSON configuration.\n\n"
            "Requires an **admin** access to the worker or its repository."
        ),
        responses={
            201: WorkerConfigurationSerializer,
            400: WorkerConfigurationExistsErrorSerializer
        },
        examples=[OpenApiExample(
            status_codes=["400"],
            response_only=True,
            name="config-exists",
            value={"id": ["01fd00d9-1020-4d84-85f2-d0785e2c5b4a"], "configuration": ["A worker configuration with this configuration already exists for this worker."]},
            description="An error where a worker configuration with this configuration already exists, including the existing worker configuration's UUID."
        )]
    ),
)
class WorkerConfigurationList(WorkerACLMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerConfigurationListSerializer
    queryset = WorkerConfiguration.objects.none()

    @cached_property
    def worker(self):
        worker = get_object_or_404(
            Worker.objects.select_related("repository"),
            pk=self.kwargs["pk"]
        )
        if self.request.method in permissions.SAFE_METHODS and not self.has_read_access(worker):
            raise PermissionDenied(detail="You do not have a guest access to this worker.")
        if self.request.method not in permissions.SAFE_METHODS and not self.has_execution_access(worker):
            raise PermissionDenied(detail="You do not have contributor access to this worker.")
        return worker

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if "pk" in self.kwargs:
            context["worker"] = self.worker
        return context

    def get_queryset(self):
        return self.worker.configurations.order_by("name")

    def filter_queryset(self, queryset):
        # Consider any value that does not look like a false, including the empty string (`?archived` alone),
        # to be true, but consider a missing parameter to be false.
        archived = self.request.query_params.get("archived", "false").lower() not in ("false", "0")
        return queryset.filter(archived=archived)


@extend_schema_view(
    get=extend_schema(description=dedent(
        """
        Retrieve details of a worker configuration.

        Requires a **read** access to the worker or its repository.
        """
    )),
    put=extend_schema(description=dedent(
        """
        Update a worker configuration.

        Requires a **write** access to the worker or its repository.
        """
    )),
    patch=extend_schema(description=dedent(
        """
        Update parts of a worker configuration.

        Requires a **write** access to the worker or its repository.
        """
    ))
)
@extend_schema(tags=["repos"])
class WorkerConfigurationRetrieve(WorkerACLMixin, RetrieveUpdateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerConfigurationSerializer

    def get_queryset(self):
        return WorkerConfiguration.objects.select_related("worker__repository")

    def check_object_permissions(self, request, obj):
        if request.method in permissions.SAFE_METHODS:
            if not self.has_read_access(obj.worker):
                raise PermissionDenied(detail="You do not have a guest access to this worker.")
        else:
            if not self.has_execution_access(obj.worker):
                raise PermissionDenied(detail="You do not have a contributor access to this worker.")

        super().check_object_permissions(request, obj)


@extend_schema(tags=["process"])
@extend_schema_view(
    get=extend_schema(description=dedent("""
        List worker runs for a given process ID.

        Requires an **admin** access on the process.
    """)),
    post=extend_schema(description=dedent("""
        Create a worker run for a given process ID.

        Requires an **admin** access to the process and a **contributor** access to the worker.
        WorkerRuns may only be created on `dataset` and `workers` processes that have not already started.
    """)),
)
class WorkerRunList(ProcessACLMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunSerializer
    queryset = WorkerRun.objects.none()

    @cached_property
    def process(self):
        process = get_object_or_404(
            Process
            .objects
            .exclude(corpus_id=None)
            .select_related("corpus")
            .annotate(has_tasks=Exists(Task.objects.filter(process=OuterRef("pk")))),
            pk=self.kwargs["pk"]
        )

        access_level = self.process_access_level(process)
        if access_level is None or access_level < Role.Admin.value:
            raise PermissionDenied(detail="You do not have an admin access to this process.")

        return process

    def get_queryset(self):
        return self.process.worker_runs \
            .select_related(
                "version__worker__repository",
                "version__worker__type",
                "version__revision__repo",
                "model_version__model",
            ) \
            .prefetch_related(Prefetch(
                "version__revision__refs",
                queryset=GitRef.objects.select_related("repository")
            )) \
            .order_by("id")

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["process"] = self.process
        return context


@extend_schema(tags=["process"])
@extend_schema_view(
    post=extend_schema(
        operation_id="CreateUserWorkerRun",
        description="Create a worker run tied to the user's local process.",
        responses={
            201: WorkerRunSerializer,
            400: OpenApiResponse(
                response=inline_serializer(
                    name="Error message with ID",
                    fields={
                        "detail": serializers.ListField(child=serializers.CharField(), required=False),
                        "id": serializers.UUIDField(required=False, help_text="ID of an existing worker run")
                    }
                ),
                description="If a local worker run with the same worker version, model version and "
                            "configuration already exists, then its UUID is returned.",
            )
        },
    ),
)
class UserWorkerRunCreate(ProcessACLMixin, CreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = UserWorkerRunSerializer
    queryset = WorkerRun.objects.none()

    @cached_property
    def local_process(self):
        try:
            return Process.objects.annotate(last_run=Value(None, output_field=IntegerField())).get(
                creator=self.request.user,
                mode=ProcessMode.Local
            )
        except Process.DoesNotExist:
            return None

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["local_process"] = self.local_process
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        run = serializer.instance
        return Response(
            status=status.HTTP_201_CREATED,
            data=WorkerRunSerializer(run, context=self.get_serializer_context()).data,
        )


@extend_schema(tags=["process"])
@extend_schema_view(
    get=extend_schema(
        operation_id="ListUserWorkerRuns",
        description=dedent("""
        List a user's worker runs tied to local processes.
    """)
    ),
)
class UserWorkerRunList(ProcessACLMixin, ListAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunSerializer
    queryset = WorkerRun.objects.none()

    @cached_property
    def local_process(self):
        try:
            local_process = Process.objects.annotate(last_run=Value(None, output_field=IntegerField())).get(
                creator=self.request.user,
                mode=ProcessMode.Local
            )
        except Process.DoesNotExist:
            local_process = None

        return local_process

    def get_queryset(self):
        if not self.local_process:
            return []
        return self.local_process.worker_runs \
            .select_related(
                "version__worker__repository",
                "version__worker__type",
                "version__revision__repo",
                "configuration",
                "model_version__model",
            ) \
            .prefetch_related(Prefetch(
                "version__revision__refs",
                queryset=GitRef.objects.select_related("repository")
            )) \
            .order_by("-created")


@extend_schema(tags=["process"])
@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveWorkerRun",
        description=dedent("""
            Retrieve details of a worker run.

            Requires an **admin** access on the process or authentication as a Ponos agent.

            WorkerRuns may only be retrieved on `dataset` and `workers` processes, or on the `local` process of the authenticated user.

            Agents may only retrieve worker runs on processes that have tasks assigned to them.
        """),
    ),
    patch=extend_schema(
        operation_id="PartialUpdateWorkerRun",
        description=dedent("""
            Partially update a worker run.

            Requires an **admin** access on the process.

            WorkerRuns may only be updated on `dataset` and `workers` processes that have not already started.
        """),
    ),
    put=extend_schema(
        operation_id="UpdateWorkerRun",
        description=dedent("""
            Update a worker run.

            Requires an **admin** access on the process.

            WorkerRuns may only be updated on `dataset` and `workers` processes that have not already started.
        """),
    ),
    delete=extend_schema(
        operation_id="DestroyWorkerRun",
        description=dedent("""
            Delete a worker run.

            Requires an **admin** access on the process.

            WorkerRuns may only be deleted from `dataset` and `workers` processes that have not already started.
        """),
    ),
)
class WorkerRunDetails(ProcessACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve or edit a worker run i.e. update its dependencies.
    Requires an admin access to the process project.
    """
    permission_classes = (IsVerifiedOrAgent, )
    serializer_class = WorkerRunEditSerializer

    def get_object(self):
        if not hasattr(self, "_worker_run"):
            self._worker_run = super().get_object()

            # Require admin access to the process, except for the user's own local process and Ponos agents.
            # For agents, we only have to check the HTTP method since the query filters handle the access rights.
            if getattr(self.request.user, "is_agent", False):
                if self.request.method not in permissions.SAFE_METHODS:
                    raise PermissionDenied(detail="You do not have an admin access to this process.")

            elif self._worker_run.process.mode != ProcessMode.Local or self._worker_run.process.creator != self.request.user:
                access_level = self.process_access_level(self._worker_run.process)
                if access_level is None or access_level < Role.Admin.value:
                    raise PermissionDenied(detail="You do not have an admin access to this process.")

            # Django does not support annotations on nested objects, so we apply it manually
            self._worker_run.process.has_tasks = self._worker_run.process_has_tasks

        return self._worker_run

    def get_queryset(self):
        if getattr(self.request.user, "is_agent", False):
            # Ponos agents only get access to all runs of any process with tasks assigned to this agent.
            # There is no link between WorkerRuns and Tasks so we cannot get more specific.
            run_filters = Exists(Task.objects.filter(process=OuterRef("process_id"), agent=self.request.user))
        else:
            run_filters = (
                # Ignore processes without corpora
                ~Q(process__corpus_id=None)
                # Except for the user's local process
                | Q(process__creator_id=self.user.id, process__mode=ProcessMode.Local)
            )

        queryset = WorkerRun.objects \
            .filter(run_filters) \
            .select_related("process__corpus") \
            .annotate(process_has_tasks=Exists(Task.objects.filter(process=OuterRef("process_id"))))

        if self.request.method.lower() != "delete":
            # Selecting those extra tables is only necessary when the request will return the serialized WorkerRun
            queryset = (
                queryset
                .select_related(
                    "version__worker__type",
                    "version__revision__repo",
                    "configuration",
                    "model_version__model",
                )
                .prefetch_related(Prefetch(
                    "version__revision__refs",
                    queryset=GitRef.objects.select_related("repository")
                ))
            )

        return queryset

    def perform_destroy(self, instance):
        # No serializer is used when destroying, so we validate the process mode here instead.
        errors = []
        if instance.process.mode not in (ProcessMode.Workers, ProcessMode.Dataset):
            errors.append("WorkerRuns can only be deleted from Workers or Dataset processes.")
        if instance.process.has_tasks:
            errors.append("WorkerRuns cannot be deleted from a process that has already started.")
        if errors:
            raise ValidationError(errors)

        # Remove references to this instance from its children parents field
        instance.process.worker_runs.filter(parents__contains=[instance.id]).update(parents=ArrayRemove("parents", instance.id))

        return super().perform_destroy(instance)


@extend_schema_view(get=extend_schema(
    operation_id="ListProcessElements",
    parameters=[
        OpenApiParameter(
            "with_image",
            type=bool,
            default=False,
            description="Include zone and image information in the response.",
        )
    ],
    responses=PolymorphicProxySerializer(
        component_name="ListProcessElementsResponse",
        serializers={
            True: ProcessElementSerializer,
            False: ProcessElementLightSerializer,
        },
        resource_type_field_name="with_image",
    ),
    tags=["process"]
))
class ListProcessElements(CorpusACLMixin, ListAPIView):
    """
    List all elements for a process with workers.\n\n
    Requires an **admin** access to the process corpus.
    """
    pagination_class = CountCursorPagination
    permission_classes = (IsVerified, )
    # For OpenAPI type discovery
    queryset = Element.objects.none()

    @property
    def with_image(self):
        return self.request.query_params.get("with_image", "false").lower() in ("true", "1")

    @cached_property
    def process(self):
        return get_object_or_404(
            # Avoid stale read on newly created processes
            Process.objects.using("default").filter(corpus_id__isnull=False),
            Q(pk=self.kwargs["pk"])
        )

    def get_serializer_class(self):
        if self.with_image:
            return ProcessElementSerializer
        else:
            return ProcessElementLightSerializer

    def get_queryset(self):
        if not self.has_admin_access(self.process.corpus):
            raise PermissionDenied(detail="You do not have an admin access to the corpus of this process.")

        try:
            queryset = self.process.list_elements()
        except AssertionError as e:
            raise ValidationError({"__all__": [str(e)]})

        if not self.with_image:
            return queryset.values("id", "type_id", "name", "confidence")

        return queryset.annotate(image_url=annotate_image_url()).values(
            "id",
            "type_id",
            "name",
            "confidence",
            "image_id",
            "image__width",
            "image__height",
            "polygon",
            "image_url",
            "rotation_angle",
            "mirrored",
        )


class UpdateWorkerActivity(GenericAPIView):
    """
    Allow a Ponos task to update an element's state
    Only allow defined evolutions of the element's state
    """
    authentication_classes = (TaskAuthentication, )
    permission_classes = (IsTask, )
    serializer_class = WorkerActivitySerializer
    queryset = WorkerActivity.objects.none()

    @cached_property
    def allowed_transitions(self):
        # Defines a list of allowed previous states for any transition
        queued = WorkerActivityState.Queued.value
        started = WorkerActivityState.Started.value
        error = WorkerActivityState.Error.value
        processed = WorkerActivityState.Processed.value
        return {
            queued: [error],
            started: [
                queued,
                # If a worker failed for some reason to update an activity to a processed or error state,
                # retrying would cause the new worker to try to update from started to started.
                started,
                error
            ],
            error: [started],
            processed: [started],
        }

    @extend_schema(
        tags=["ml"],
        responses={
            200: WorkerActivitySerializer,
            409: None,
        },
        operation_id="UpdateWorkerActivity",
        description=dedent("""
            Updates the activity of a WorkerRun on an element.

            Only a Ponos task authenticated via task authentication may perform this request.

            A **HTTP_409_CONFLICT** is returned in case the body is valid but the update failed.

            Updating to an `error` state will also update any other `queued` activities on the same element and process to `error`.
        """).strip(),
        parameters=[
            OpenApiParameter(
                "id",
                type=UUID,
                location=OpenApiParameter.PATH,
                description=dedent("""
                    ID of a WorkerRun or of a WorkerVersion to update an activity for.

                    WorkerVersion IDs may only be used when there is only one WorkerRun
                    on the specified process that uses this WorkerVersion.

                    Using a WorkerVersion ID is deprecated.
                """),
                required=True,
            ),
        ],
    )
    def put(self, request, *args, **kwarg):
        """
        Update a worker activity with two database requests.
        If no existing activity is found, the update count is 0.
        If the new state is disallowed, the update count is 0 too.
        """
        serializer = self.get_serializer(data=request.data, partial=False)
        serializer.is_valid(raise_exception=True)

        version_or_run_id = self.kwargs["pk"]
        element_id = serializer.validated_data["element_id"]
        state = serializer.validated_data["state"].value
        process_id = serializer.validated_data.get("process_id")

        try:
            worker_run = WorkerRun.objects.get(id=version_or_run_id)
        except WorkerRun.DoesNotExist:
            if not process_id:
                raise ValidationError({
                    "detail": [
                        "Either this WorkerRun ID does not exist, "
                        "or a WorkerVersion ID was sent without a `process_id`."
                    ]
                })

            # Retrieve up to two WorkerRuns on the process with this worker version ID.
            # If there are two WorkerRuns, we cannot properly tell which activities to update,
            # so we abort early on and invite to use WorkerRun IDs instead of WorkerVersion IDs.
            # We should not raise HTTP 409, since it has special handling in base-worker
            # that would make this message invisible to users.
            worker_runs = list(WorkerRun.objects.filter(process_id=process_id, version_id=version_or_run_id)[:2])
            if len(worker_runs) > 1:
                raise ValidationError({
                    "detail": [
                        "There is more than one WorkerRun on this process with this WorkerVersion. "
                        "Please update the worker to call UpdateWorkerActivity with WorkerRun IDs, "
                        "run the duplicated WorkerVersion in a process without worker activities, "
                        "or run each duplicated WorkerVersion in a separate process."
                    ]
                })

            worker_run = worker_runs[0] if worker_runs else None

        worker_version_id, configuration_id, model_version_id = None, None, None
        if worker_run:
            worker_version_id = worker_run.version_id
            configuration_id = worker_run.configuration_id
            model_version_id = worker_run.model_version_id
            process_id = worker_run.process_id

        # Between zero and one worker activity can match the filter due to the DB constraint
        activity_filter = Q(
            worker_version_id=worker_version_id,
            element_id=element_id,
            configuration_id=configuration_id,
            model_version_id=model_version_id,
            process_id=process_id,
            state__in=self.allowed_transitions[state]
        )

        fields_to_update = {"state": state}

        # If we are switching to started, we need to restrict to activities that either are not started,
        # or have not been updated for an hour, and also update the `started` timestamp.
        if state == WorkerActivityState.Started.value:
            activity_filter &= (
                ~Q(state=WorkerActivityState.Started)
                | Q(updated__lte=Now() - timedelta(seconds=settings.WORKER_ACTIVITY_TIMEOUT))
            )
            fields_to_update["started"] = Now()

        activity = WorkerActivity.objects.filter(activity_filter)

        update_count = activity.update(**fields_to_update)

        if not update_count:
            # As no row has been updated the provided data was in conflict with the actual state
            raise ConflictAPIException(
                {
                    "__all__": [
                        "Either this worker activity does not exist, "
                        "is assigned to another process, "
                        f"or updating the state to '{state}' is forbidden."
                    ]
                }
            )

        # Update other activities
        if state == WorkerActivityState.Error.value:
            WorkerActivity.objects.filter(
                process_id=process_id,
                element_id=element_id,
                state=WorkerActivityState.Queued,
            ).update(state=WorkerActivityState.Error)

        return Response(serializer.data)


class WorkerActivityBase(ListAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerStatisticsSerializer
    pagination_class = None
    queryset = WorkerActivity.objects \
        .values(
            "worker_version_id",
            "configuration_id",
            "model_version_id"
        ).annotate(
            average_processed_time=Avg(
                F("updated") - F("started"),
                filter=Q(state=WorkerActivityState.Processed, started__isnull=False),
            ),
            max_processed_time=Max(
                F("updated") - F("started"),
                filter=Q(state=WorkerActivityState.Processed, started__isnull=False),
            ),
            min_processed_time=Min(
                F("updated") - F("started"),
                filter=Q(state=WorkerActivityState.Processed, started__isnull=False),
            ),
            median_processed_time=PercentileCont(
                0.5,
                order_by=F("updated") - F("started"),
                filter=Q(state=WorkerActivityState.Processed, started__isnull=False),
                output_field=DurationField(),
            ),
        ) \
        .annotate(
            # Add the counts per state in a separate annotate() call, because otherwise the F('started')
            # on the processed time annotations would be interpreted as being the started activities count,
            # not the started datetime, causing an error: "'<CombinedExpression: F(updated) - F(started)>' is an aggregate"
            **{
                state.value: Count("id", filter=Q(state=state.value))
                for state in WorkerActivityState
            }
        ) \
        .annotate(
            # Completion time estimate based on median time per task from the previously annotated data
            completion_estimate_time=ExpressionWrapper(
                (
                    F("median_processed_time") * (
                        F(WorkerActivityState.Queued.value)
                        + F(WorkerActivityState.Started.value)
                    ) / F("process__chunks")
                ),
                output_field=DurationField(),
            )
        )


@extend_schema_view(
    get=extend_schema(
        operation_id="CorpusWorkersActivity",
        tags=["process"]
    )
)
class CorpusWorkersActivity(CorpusACLMixin, WorkerActivityBase):
    """
    Retrieve corpus wise statistics about the activity of all its worker processes.\n
    Requires a **guest** access.
    """

    def filter_queryset(self, queryset):
        corpus = self.get_corpus(self.kwargs["corpus"], role=Role.Guest)
        return queryset.filter(element_id__in=corpus.elements.values("id"))


@extend_schema_view(
    get=extend_schema(
        operation_id="ProcessWorkersActivity",
        tags=["process"]
    )
)
class ProcessWorkersActivity(ProcessACLMixin, WorkerActivityBase):
    """
    Retrieve process statistics about the activity of its workers.\n
    Requires a **guest** access.
    """

    def filter_queryset(self, queryset):
        process = get_object_or_404(Process.objects.all(), pk=self.kwargs["pk"])
        access_level = self.process_access_level(process)
        if not access_level:
            raise NotFound
        if access_level < Role.Guest.value:
            raise PermissionDenied(detail="You do not have a guest access to this process.")

        return queryset.filter(process_id=process.id)


@extend_schema_view(
    get=extend_schema(
        operation_id="ListWorkerActivities",
        tags=["process"],
        parameters=[
            OpenApiParameter(
                "process_id",
                type=UUID,
                description="Filter worker activities by process ID",
                required=False,
            ),
            OpenApiParameter(
                "state",
                enum=[state.value for state in WorkerActivityState],
                description="Filter worker activities by state",
                required=False,
            ),
        ]
    )
)
class WorkerActivityList(CorpusACLMixin, ProcessACLMixin, ListAPIView):
    """
    List worker activities in a corpus.

    Requires guest access to the corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerActivitySerializer

    @cached_property
    def corpus(self):
        return self.get_corpus(self.kwargs["corpus"])

    def get_queryset(self):
        return WorkerActivity.objects.filter(element__corpus=self.corpus).order_by("id")

    def filter_queryset(self, queryset):
        errors = {}

        process_id = None
        if "process_id" in self.request.query_params:
            try:
                process_id = UUID(self.request.query_params["process_id"])
            except (TypeError, ValueError):
                errors["process_id"] = ["Process ID should be an UUID."]

        if process_id:
            try:
                # This uses a select_related because the process access level check looks for the corpus
                process = Process.objects.select_related("corpus").get(id=process_id, corpus=self.corpus)
            except Process.DoesNotExist:
                errors["process_id"] = ["This process does not exist."]
            else:
                access_level = self.process_access_level(process)
                if not access_level or access_level < Role.Guest.value:
                    errors["process_id"] = ["You do not have a guest access to this process."]
                queryset = queryset.filter(process=process)

        if "state" in self.request.query_params:
            try:
                state = WorkerActivityState(self.request.query_params["state"])
            except (TypeError, ValueError):
                errors["state"] = [f"{self.request.query_params['state']!r} is not a valid WorkerActivity state"]
            else:
                queryset = queryset.filter(state=state)

        if errors:
            raise ValidationError(errors)

        return queryset


@extend_schema_view(post=extend_schema(
    operation_id="CreateProcessTemplate",
    responses={201: ProcessSerializer},
    tags=["process"]
))
class CreateProcessTemplate(ProcessACLMixin, WorkerACLMixin, CreateAPIView):
    """
    Create a `template` process from an existing `dataset` or `workers` process,
    to copy its worker runs to other processes.

    Requires a **contributor** access to the existing process and to each worker of the process.
    """
    permission_classes = (IsVerified, )
    serializer_class = CreateProcessTemplateSerializer

    def get_queryset(self):
        return (
            Process
            .objects
            .prefetch_related(Prefetch("worker_runs", queryset=WorkerRun.objects.select_related(
                "version__worker__type",
                "version__worker__repository",
            )))
            .select_related("corpus")
        )

    def check_object_permissions(self, request, process):
        access_level = self.process_access_level(process)
        if not access_level or access_level < Role.Contributor.value:
            raise PermissionDenied(detail="You do not have a contributor access to this process.")

        errors = []

        if process.mode not in (ProcessMode.Dataset, ProcessMode.Workers):
            errors.append("Templates can only be created from Workers or Dataset processes.")

        if not process.worker_runs.all():
            errors.append("This process does not have any worker runs.")

        if not all(self.has_execution_access(worker_run.version.worker) for worker_run in process.worker_runs.all()):
            errors.append("You do not have an execution access to every worker of this process.")

        if errors:
            raise ValidationError(errors)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["template"] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(
            status=status.HTTP_201_CREATED,
            data=ProcessSerializer(serializer.instance, context={"request": self.request}).data, headers=headers
        )


@extend_schema_view(post=extend_schema(
    operation_id="ApplyProcessTemplate",
    parameters=[
        OpenApiParameter(
            "id",
            type=UUID,
            location=OpenApiParameter.PATH,
            description="ID of the template to apply",
            required=True,
        ),
    ],
    responses={200: ProcessSerializer, 403: None},
    tags=["process"],
))
class ApplyProcessTemplate(ProcessACLMixin, WorkerACLMixin, CreateAPIView):
    """
    Apply a template on a given process.

    Requires a **Contributor** access to both the template and the target process. The target process must have no previously applied template.
    """
    permission_classes = (IsVerified, )
    # For OpenAPI type discovery
    serializer_class = ApplyProcessTemplateSerializer

    def get_queryset(self):
        return (
            Process.objects
            .filter(mode=ProcessMode.Template)
            .prefetch_related(Prefetch("worker_runs", queryset=WorkerRun.objects.select_related(
                "version__worker__type",
                "version__worker__repository",
                "model_version__model",
            )))
            .select_related("corpus")
        )

    def check_object_permissions(self, request, template):
        access_level = self.process_access_level(template)
        if not access_level or access_level < Role.Contributor.value:
            raise PermissionDenied(detail="You do not have a contributor access to this process.")

        for worker_run in template.worker_runs.all():
            if not self.has_execution_access(worker_run.version.worker):
                raise PermissionDenied(detail="You do not have an execution access to this worker.")

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["template"] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data, user=request.user)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        process = serializer.instance
        return Response(
            status=status.HTTP_200_OK,
            data=ProcessSerializer(process, context={"request": self.request}).data
        )


@extend_schema_view(
    delete=extend_schema(
        operation_id="ClearProcess",
        tags=["process"],
        responses={
            204: None,
        },
    )
)
class ClearProcess(ProcessACLMixin, DestroyAPIView):
    """
    Remove all templates and/or workers from a (not started) process
    """
    permission_classes = (IsVerified, )
    queryset = Process.objects.all().annotate(has_tasks=Exists(Task.objects.filter(process=OuterRef("pk"))))

    def check_object_permissions(self, request, process):
        super().check_object_permissions(request, process)

        access_level = self.process_access_level(process)
        required_access = Role.Admin.value

        if not access_level:
            raise NotFound
        if access_level < required_access:
            raise PermissionDenied(detail="You do not have a sufficient access level to this process.")
        if process.has_tasks:
            raise ValidationError({"__all__": ["A process can only be cleared before getting started."]})

    def destroy(self, request, *args, **kwargs):
        process = self.get_object()
        process.clear()
        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(tags=["ingest"])
class BucketList(ListAPIView):
    """
    List the S3 buckets that are available for ingestion.
    """

    permission_classes = (IsVerified, )
    scopes = (Scope.S3Ingest, )
    pagination_class = None
    serializer_class = BucketSerializer

    def get_queryset(self):
        resource = get_ingest_resource()
        # Listing buckets does not use any pagination at all,
        # so this will make a single API query once.
        buckets = list(resource.buckets.all())

        # Add fake Bucket instances if any extra buckets were specified in the settings
        # Note that duplicate bucket names are not checked
        if settings.INGEST_EXTRA_BUCKETS:
            buckets += [
                resource.Bucket(name=name)
                for name in settings.INGEST_EXTRA_BUCKETS
            ]

        # Most S3 API implementations sort buckets by name,
        # but that is not specified anywhere, so we sort manually
        return sorted(buckets, key=lambda bucket: bucket.name)


@extend_schema(tags=["ingest"])
class S3ImportCreate(CreateAPIView):
    """
    Create a new S3 import.

    Requires a verified email and the `s3_ingest` scope.
    """

    permission_classes = (IsVerified, )
    scopes = (Scope.S3Ingest, )
    serializer_class = S3ImportSerializer


@extend_schema_view(
    post=extend_schema(
        operation_id="SelectProcessFailures",
        tags=["process"],
        responses={204: None},
    ),
)
class SelectProcessFailures(ProcessACLMixin, CreateAPIView):
    """
    Add elements with a WorkerActivity in an error state to user's selection
    """
    permission_classes = (IsVerified, )
    # Empty serializer as no parameters are passed yet
    serializer_class = Serializer

    def create(self, *args, **kwargs):
        super().create(*args, **kwargs)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_create(self, serializer):
        """
        Select elements of a process that have any worker activity failure.
        The process must be finished and of type "workers".
        """
        # TODO: Move this to the `queryset` class attribute and make this endpoint more in line with best practices
        qs = (
            Process.objects
            .select_related("corpus")
            .filter(
                mode=ProcessMode.Workers,
                corpus_id__isnull=False,
            )
            .annotate(has_non_final_tasks=Exists(
                Task
                .objects
                .filter(process=OuterRef("pk"))
                .exclude(state__in=FINAL_STATES)
            ))
        )
        process = get_object_or_404(qs, pk=self.kwargs["pk"])
        access_level = self.process_access_level(process)
        if access_level is None or access_level < Role.Guest.value:
            raise PermissionDenied(detail="You do not have a read access to the corpus of this process.")
        if process.activity_state is not ActivityState.Ready:
            raise ValidationError({
                "__all__": ["The process must use worker activities to select elements with failures."]
            })
        if process.has_non_final_tasks:
            raise ValidationError({
                "__all__": ["The process must be finished to select elements with failures."]
            })

        elt_ids = WorkerActivity.objects.filter(
            state=WorkerActivityState.Error,
            process=process,
        ).distinct("element_id").values_list("element_id", flat=True)
        if not elt_ids.exists():
            raise ValidationError({
                "__all__": ["This process has no element in a failure state."]
            })
        # This method could be optimised or run as an asynchronous task but
        # usually processes have few elements with a worker activity failure
        Selection.objects.bulk_create(
            (
                Selection(user=self.request.user, element_id=elt_id) for elt_id in elt_ids.iterator()
            ),
            ignore_conflicts=True,
        )


@extend_schema_view(
    post=extend_schema(
        responses={201: WorkerVersionSerializer},
        tags=["process"],
    )
)
class CreateDockerWorkerVersion(CreateAPIView):
    """
    Create a worker version tied to a remote docker image. The worker version is directly tagged as available.

    Causes an HTTP 400 error if the worker version already exists, unless a new Git reference has been
    added to the existing worker version.

    Cannot be used with Ponos task or agent authentication. Requires the `CreateDockerWorkerVersion` scope.
    """
    permission_classes = (IsVerified, )
    serializer_class = DockerWorkerVersionSerializer
    # Only user accounts can call that endpoint
    authentication_classes = (SessionAuthentication, TokenAuthentication)
    scopes = (Scope.CreateDockerWorkerVersion,)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        return Response(
            status=status.HTTP_201_CREATED,
            data=WorkerVersionSerializer(serializer.instance, context={"request": self.request}).data,
        )
