import json
from datetime import datetime, timezone
from unittest.mock import call, patch

from rest_framework import status
from rest_framework.reverse import reverse

from arkindex.documents.models import Corpus
from arkindex.process.models import ProcessMode, WorkerConfiguration, WorkerRun, WorkerVersion, WorkerVersionState
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState
from arkindex.users.models import Role, User


class TestTemplates(FixtureAPITestCase):
    """
    Test workflows from Arkindex existing elements
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name="private")
        cls.private_corpus.elements.create(
            type=cls.private_corpus.types.create(slug="volume", folder=True),
            name="Hidden",
        )
        cls.private_process_template = cls.private_corpus.processes.create(
            creator=cls.user, mode=ProcessMode.Workers
        )
        cls.private_template = cls.private_corpus.processes.create(
            creator=cls.user, mode=ProcessMode.Template
        )
        # Workers ProcessMode
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_1 = cls.version_1.worker
        cls.worker_1.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.version_2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.worker_2 = cls.version_2.worker
        cls.worker_2.memberships.create(user=cls.user, level=Role.Contributor.value)

        cls.process = cls.corpus.processes.create(
            creator=cls.user, mode=ProcessMode.Workers
        )
        cls.process_template = cls.corpus.processes.create(
            creator=cls.user, mode=ProcessMode.Workers
        )
        cls.template = cls.corpus.processes.create(
            creator=cls.user, mode=ProcessMode.Template
        )
        cls.worker_configuration = WorkerConfiguration.objects.create(
            worker=cls.version_1.worker,
            name="A config",
            configuration={"param1": "value1"},
        )
        run_1 = cls.process_template.worker_runs.create(
            version=cls.version_1, parents=[], configuration=cls.worker_configuration
        )
        cls.process_template.worker_runs.create(
            version=cls.version_2,
            parents=[run_1.id],
        )

        cls.model = Model.objects.create(name="moo")
        cls.model_version = cls.model.versions.create(state=ModelVersionState.Available)

        run_1 = cls.template.worker_runs.create(
            version=cls.version_1, parents=[], configuration=cls.worker_configuration
        )
        cls.template.worker_runs.create(
            version=cls.version_2,
            parents=[run_1.id],
            model_version=cls.model_version,
        )

        cls.private_process_template.worker_runs.create(
            version=cls.version_1, parents=[], configuration=cls.worker_configuration
        )
        cls.private_template.worker_runs.create(
            version=cls.version_1, parents=[], configuration=cls.worker_configuration
        )

    def test_create(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse(
                    "api:create-process-template", kwargs={"pk": str(self.process_template.id)}
                ),
                {"name": "test_template"},
            )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["mode"], "template")

        child_run, parent_run = WorkerRun.objects.select_related("version__worker").filter(process__id=response.json()["id"]).order_by("version__worker__slug").all()
        # Check dependency
        self.assertListEqual(child_run.parents, [parent_run.id])

        # Check that every new worker_run is the same as one of the template's
        self.assertTrue(self.process_template.worker_runs.filter(version=parent_run.version).exists())
        self.assertTrue(self.process_template.worker_runs.filter(version=child_run.version).exists())

    def test_create_requires_authentication(self):
        response = self.client.post(
            reverse("api:create-process-template", kwargs={"pk": str(self.process_template.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(), {"detail": "Authentication credentials were not provided."}
        )

    def test_create_requires_worker_runs(self):
        process_no_worker_runs = self.corpus.processes.create(
            creator=self.user, mode=ProcessMode.Workers
        )
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:create-process-template", kwargs={"pk": str(process_no_worker_runs.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(
            response.json(),
            ["This process does not have any worker runs."],
        )

    def test_create_requires_verified_user(self):
        new_user = User.objects.create(email="new@test.fr", verified_email=False)
        self.client.force_login(new_user)
        response = self.client.post(
            reverse("api:create-process-template", kwargs={"pk": str(self.process_template.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Guest.value)
    def test_create_requires_contributor_access_rights_process(self, get_max_level_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse(
                    "api:create-process-template",
                    kwargs={"pk": str(self.private_process_template.id)},
                )
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have a contributor access to this process."},
        )
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.private_corpus))

    @patch("arkindex.project.mixins.has_access", return_value=False)
    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Contributor.value)
    def test_create_requires_access_rights_all_workers(self, get_max_level_mock, has_access_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:create-process-template", kwargs={"pk": str(self.private_process_template.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            response.json(),
            ["You do not have an execution access to every worker of this process."],
        )
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.private_corpus))
        self.assertEqual(has_access_mock.call_args_list, [
            call(self.user, self.worker_1, Role.Contributor.value, skip_public=False),
            call(self.user, self.worker_1.repository, Role.Contributor.value, skip_public=False),
        ])

    def test_create_unsupported_mode(self):
        self.client.force_login(self.user)
        for mode in set(ProcessMode) - {ProcessMode.Workers, ProcessMode.Dataset, ProcessMode.Local}:
            with self.subTest(mode=mode):
                self.process.mode = mode
                self.process.save()

                with self.assertNumQueries(4):
                    response = self.client.post(
                        reverse("api:create-process-template", kwargs={"pk": str(self.template.id)}),
                        data=json.dumps({"process_id": str(self.process.id)}),
                        content_type="application/json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), ["Templates can only be created from Workers or Dataset processes."])
                self.process.refresh_from_db()
                self.assertEqual(self.process.template, None)

    def test_create_local(self):
        self.client.force_login(self.user)
        local_process = self.user.processes.get(mode=ProcessMode.Local)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:create-process-template", kwargs={"pk": str(self.template.id)}),
                data=json.dumps({"process_id": str(local_process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), ["Templates can only be created from Workers or Dataset processes."])
        local_process.refresh_from_db()
        self.assertEqual(local_process.template, None)

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Guest.value)
    def test_apply_requires_contributor_rights_on_template(self, get_max_level_mock):
        """Raise 403 if the user does not have rights on template
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.private_template.id)}),
                data=json.dumps({"process_id": str(self.process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have a contributor access to this process."},
        )
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.private_corpus))

    @patch("arkindex.project.mixins.get_max_level", side_effect=[Role.Contributor.value, Role.Guest.value])
    def test_apply_requires_contributor_rights_on_process(self, get_max_level_mock):
        """Raise 403 if the user does not have rights on the target process
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.private_template.id)}),
                data=json.dumps({"process_id": str(self.process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have a contributor access to this process."},
        )
        self.assertEqual(get_max_level_mock.call_args_list, [
            call(self.user, self.private_corpus),
            call(self.user, self.corpus),
        ])

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_apply_requires_access_rights_all_workers(self, has_access_mock):
        """Raise 403 if the user does not have rights on all workers concerned
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.private_template.id)}),
                data=json.dumps({"process_id": str(self.process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have an execution access to this worker."},
        )
        self.assertEqual(has_access_mock.call_args_list, [
            call(self.user, self.worker_1, Role.Contributor.value, skip_public=False),
            call(self.user, self.worker_1.repository, Role.Contributor.value, skip_public=False),
        ])

    def test_apply_already_applied(self):
        """Raise 400 if the process already has a template attached
        """
        process_with_template = self.corpus.processes.create(
            creator=self.user, mode=ProcessMode.Workers, template=self.template
        )
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
            data=json.dumps({"process_id": str(process_with_template.id)}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"process_id": ["A template has already been applied to this process."]},
        )

    def test_apply_without_a_template(self):
        """Raise 404 if the template is not a template
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:apply-process-template", kwargs={"pk": str(self.process.id)}),
            data=json.dumps({"process_id": str(self.process.id)}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_apply(self):
        self.assertIsNotNone(self.version_2.docker_image_iid)
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data=json.dumps({"process_id": str(self.process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["template_id"], str(self.template.id))

        child_run, parent_run = WorkerRun.objects.select_related("version__worker").filter(process__id=response.json()["id"]).order_by("version__worker__slug").all()

        self.assertEqual(parent_run.process_id, self.process.id)
        self.assertEqual(parent_run.version_id, self.version_1.id)
        self.assertIsNone(parent_run.model_version_id)
        self.assertEqual(parent_run.configuration_id, self.worker_configuration.id)
        self.assertListEqual(parent_run.parents, [])

        self.assertEqual(child_run.process_id, self.process.id)
        self.assertEqual(child_run.version_id, self.version_2.id)
        self.assertEqual(child_run.model_version_id, self.model_version.id)
        self.assertIsNone(child_run.configuration_id)
        self.assertListEqual(child_run.parents, [parent_run.id])

    def test_apply_docker_image_iid(self):
        """
        Either docker_image_id or docker_image_iid attribute can be set
        """

        self.version_2.docker_image_id = None
        self.version_2.docker_image_iid = "test"
        self.version_2.save()

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data=json.dumps({"process_id": str(self.process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["template_id"], str(self.template.id))

    def test_apply_delete_previous_worker_runs(self):
        self.client.force_login(self.user)
        # Create a process with one worker run already
        process = self.corpus.processes.create(
            creator=self.user, mode=ProcessMode.Workers
        )
        process.worker_runs.create(
            version=self.version_2,
            parents=[],
        )
        # Apply a template that has two other worker runs
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data=json.dumps({"process_id": str(process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["template_id"], str(self.template.id))

        # Assert that the previous worker runs was deleted and the template was correctly applied
        child_run, parent_run = WorkerRun.objects.select_related("version__worker").filter(process__id=response.json()["id"]).order_by("version__worker__slug").all()

        self.assertEqual(parent_run.process_id, process.id)
        self.assertEqual(parent_run.version_id, self.version_1.id)
        self.assertIsNone(parent_run.model_version_id)
        self.assertEqual(parent_run.configuration_id, self.worker_configuration.id)
        self.assertListEqual(parent_run.parents, [])

        self.assertEqual(child_run.process_id, process.id)
        self.assertEqual(child_run.version_id, self.version_2.id)
        self.assertEqual(child_run.model_version_id, self.model_version.id)
        self.assertIsNone(child_run.configuration_id)
        self.assertListEqual(child_run.parents, [parent_run.id])

    def test_apply_unavailable_worker_version(self):
        self.version_2.state = WorkerVersionState.Error
        self.version_2.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data=json.dumps({"process_id": str(self.process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["This template contains one or more unavailable worker versions and cannot be applied."]})
        self.process.refresh_from_db()
        self.assertEqual(self.process.template, None)

    def test_apply_archived_worker(self):
        self.worker_2.archived = datetime.now(timezone.utc)
        self.worker_2.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data=json.dumps({"process_id": str(self.process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["This template contains one or more worker versions from archived workers and cannot be applied."]})
        self.process.refresh_from_db()
        self.assertEqual(self.process.template, None)

    def test_apply_unavailable_model_version(self):
        self.model_version.state = ModelVersionState.Error
        self.model_version.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data=json.dumps({"process_id": str(self.process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"non_field_errors": ["This template contains one or more unavailable model versions and cannot be applied."]})

        self.process.refresh_from_db()
        self.assertEqual(self.process.template, None)

    def test_apply_archived_model(self):
        self.model.archived = datetime.now(timezone.utc)
        self.model.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data=json.dumps({"process_id": str(self.process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"non_field_errors": ["This template contains one or more model versions from archived models and cannot be applied."]})
        self.process.refresh_from_db()
        self.assertEqual(self.process.template, None)

    def test_apply_unsupported_mode(self):
        self.client.force_login(self.user)
        for mode in set(ProcessMode) - {ProcessMode.Workers, ProcessMode.Dataset, ProcessMode.Local}:
            with self.subTest(mode=mode):
                self.process.mode = mode
                self.process.save()

                with self.assertNumQueries(5):
                    response = self.client.post(
                        reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                        data=json.dumps({"process_id": str(self.process.id)}),
                        content_type="application/json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "process_id": ["Templates can only be applied on Workers or Dataset processes."],
                })
                self.process.refresh_from_db()
                self.assertEqual(self.process.template, None)

    def test_apply_local(self):
        self.client.force_login(self.user)
        local_process = self.user.processes.get(mode=ProcessMode.Local)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data=json.dumps({"process_id": str(local_process.id)}),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["Process with this ID does not exist."],
        })
        local_process.refresh_from_db()
        self.assertEqual(local_process.template, None)

    def test_list_ignores_configuration_filter(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:process-list"),
                data={"mode": "template", "with_tasks": True},
                content_type="application/json",
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # The 'with_tasks' filter should be ignored and some templates should be returned
        # If it wasn't, no template are returned because none are configured
        self.assertTrue(len(response.json()) > 0)
